/*   File: gma.cc

     Encloses all parameters for a GMA-system by propagating the constraints
     given by the ODE:s. A component of a general GMA-system is given by

          x'(t) = \sum_{i=1}^M a_i \prod_{j=1}^d x_j(t)^{g_{ij}}
    
     where a_i and g_{ij} are real parameters.

     Uses the Profil/BIAS library, available from     

	  http://www.ti3.tu-harburg.de/Software/PROFILEnglisch.html

     Syntax: ./gma dir dim tol comp [err]
     Syntax: ./gma GMA  3  1e-3  3   5
             ./gma GMA  3  1e-3  3   1
             ./gma voit 4  1e-3  1   

     Author: Warwick Tucker <warwick.tucker@math.uib.no>
     Latest edit: Tue May 13 09:52:20 CEST 2008
*/

#include "utilities.h"

#undef USE_SLOPE_CONSTRAINTS
#undef USE_GENERAL_CONSTRAINTS
#define USE_SLOPE_CONSTRAINTS
#define USE_GENERAL_CONSTRAINTS

////////////////////////////////////////////////////////////////////////////
//
// Returns \prod_{i=1; i != k}^d x_i^[p_i].
//
INTERVAL semiProd(const BOX &x, const BOX &p, int dim, int k)
{
  INTERVAL prod(1);
  for (int i = 1; i <= dim; i++)
    if ( i != k ) 
      prod *= Power(x(i), p(i));
  return prod;
}

////////////////////////////////////////////////////////////////////////////
// 
// Returns \sum{i=1; i!= k}^M [a_i]*(x_1^[p_i1]*...*x_d^[p_id]).
//
INTERVAL semiSum(const BOX &x, const IM &p, int k)
{
  INTERVAL sum(0);
  int rDim = RowDimension(p);      // rDim = M.
  int cDim = ColDimension(p);      // cDim = d + 1.
  for (int i = 1; i <= rDim; i++) 
    if ( i != k ) 
      sum += p(i, cDim)*semiProd(x, Row(p,i), cDim - 1, 0);
  return sum;
}

////////////////////////////////////////////////////////////////////////////
//
// A cSet version of the logarithm. Computes cset(log(x)).
//
INTERVAL cSetLog(const INTERVAL &x)
{ 
  static INTERVAL EMPTY_SET(+1,-1); 
  if ( Sup(x) < 0.0 )              // Domain totally negative.
    return EMPTY_SET;              
  if ( Inf(x) <= 0.0 ) {           // Domain straddles zero.
    INTERVAL xHi(Sup(x)); 
    return INTERVAL(Machine::NegInfinity, Sup(Log(xHi)));
  }
  return Log(x);                   // Natural (positive) domain.
}

////////////////////////////////////////////////////////////////////////////
//
// A cSet version of division. Computes cset(x/y).
//
INTERVAL cSetDiv(const INTERVAL &x, const INTERVAL &y)
{
  static INTERVAL REAL_LINE(Machine::NegInfinity, Machine::PosInfinity);
  if ( !(0.0 <= y) )               
    return x/y;                    // Natural domain.
  return REAL_LINE;                // Division by zero.
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the (i,j):th constraint of the GMA-system.
// Here i is the component index, and we have the following encoding
// j = 0   : the forward propagation (cone condition)
// j < cDim: constraints on the g_{ij} 
// j = cDim: constraints on the a_{i} 
//
INTERVAL constraint(const IM &p, const BOX &x, const INTERVAL &s, int i, int j)
{
  static INTERVAL REAL_LINE(Machine::NegInfinity, Machine::PosInfinity);
  int xDim = Dimension(x);
  int cDim = ColDimension(p);

  if (j == 0)                      // Slope. 
    return ( semiSum(x,p,0) );
  else if (j == cDim) {            // a_i.
    INTERVAL nominator   = s - semiSum(x,p,i);
    INTERVAL denominator = semiProd(x, Row(p, i), xDim, 0);
    return cSetDiv(nominator, denominator);
  }
  else if (j < cDim) {             // g_ij.
    INTERVAL nominator   = s - semiSum(x,p,i);
    INTERVAL denominator = p(i, cDim)*semiProd(x, Row(p, i), xDim, j);
    INTERVAL quotient    = cSetDiv(nominator, denominator);
      return cSetLog(quotient)/cSetLog(x(j));
  }
  else {
    cerr << "Error (constraint): wrong index domain!" << endl;
    exit(1);
    return INTERVAL(0);
  }
}

////////////////////////////////////////////////////////////////////////////
//
// Contracts the parameter box p via a set of constraints.
//
bool constraintPropagate(IM &p, const IM &dataSlice)
{
  int dataRowDim = RowDimension(dataSlice);
  int dataColDim = ColDimension(dataSlice);

  for (int m = 1; m <= dataRowDim; m++) {
    INTERVAL s = dataSlice(m, dataColDim);
    BOX      x(dataColDim - 2);
    for (int n = 1; n <= dataColDim - 2; n++)
      x(n) = dataSlice(m, n);

#ifdef USE_SLOPE_CONSTRAINTS
    INTERVAL c = constraint(p,x,s,0,0); 
    INTERVAL t;                         
    if ( !Intersection(t, s, c) )
      return false;
    else 
      s = t;                       // Only good for interval slopes.
#endif                             // End of cone condition.
    
#ifdef USE_GENERAL_CONSTRAINTS
    for (int i = 1; i <= RowDimension(p); i++) {
      for (int j = 1; j <= ColDimension(p); j++) 
	if ( Diam(p(i,j)) > 0.0 ) {      
	  INTERVAL c = constraint(p,x,s,i,j);
	  INTERVAL t;                                        
          if ( !Intersection(t, p(i,j), c) ) 
	    return false;
          else 
	    p(i,j) = t;
	}
    }
#endif                             // End of general constraints.
  }                                // End of all data points.
  return true;
}

////////////////////////////////////////////////////////////////////////////
//
// Contracts parameterDomain via one sweep of constraint propagations. When 
// the constraints have been performed, the parameter box is bisected. This 
// procedure repeats until all parameter boxes have size < tol or have 
// been excluded due to inconsistencies.
//
void reconstructParameters(List<IM> &resultList, const IM &dataSlice, 
			   const IM &parameterDomain, double tol)
{
  int counter(0);
  List<IM> paramList; paramList += parameterDomain;
  while ( !IsEmpty(paramList) ) {
    IM p = Pop(paramList);
    counter++;
    if ( constraintPropagate(p, dataSlice) )
      if ( Diameter(p) < tol )
	resultList += p;
      else
	splitAndStore(p, paramList);
  }
  cout << "Number of examined boxes: " << counter << endl;
}

////////////////////////////////////////////////////////////////////////////
//
// p_{i,j} = g_{i,j} (for i = 1,...,M and j = 1,...,d). p_{i,d+1} = a_i. 
//
void setParameterDomain(IM &p, string System)
{
  int rDim = RowDimension(p);      // rDim = M.
  int cDim = ColDimension(p);      // cDim = d + 1.
  for (int i = 1; i <= rDim; i++) {
    for (int j = 1; j <= cDim - 1; j++) 
      if (System == "sSystem" )
	p(i,j) = Hull(-1, +1);     // The elements of g have domain [-1 ,+1].
      else
	p(i,j) = Hull(-2 , +2);    // The elements of g have domain [-2 ,+2].
    p(i, cDim) = Hull(-20,+20);    // The elements of a have domain [-20,+20].
  }
}

////////////////////////////////////////////////////////////////////////////
//
// The exact parameter values for the GMA-system of Voit et al.
//
void setTargetParameters(IM &p, string System, int comp)
{
  if ( System == "GMA") {
  if ( comp == 1 ) {
    p(1,1)=Hull(0.00); p(1,2)=Hull(-1);   p(1,3)=Hull(-1);   p(1,4)=Hull( 0.4);
    p(2,1)=Hull(0.50); p(2,2)=Hull(-0.1); p(2,3)=Hull( 0);   p(2,4)=Hull(-3); 
    p(3,1)=Hull(0.75); p(3,2)=Hull( 0);   p(3,3)=Hull(-0.2); p(3,4)=Hull(-2); 
  }
  int cDim = ColDimension(p);      // cDim = d + 1.
  if ( comp == 2 ) {
    Resize(p,2,cDim);
    p(1,1)=Hull(0.5); p(1,2)=Hull(-0.1); p(1,3)=Hull(0);  p(1,4)=Hull( 3.0); 
    p(2,1)=Hull(0.0); p(2,2)=Hull( 0.5); p(2,3)=Hull(0);  p(2,4)=Hull(-1.5); 
  }  
  if ( comp == 3 ) {
    Resize(p,2,cDim);
    p(1,1)=Hull(0.75); p(1,2)=Hull(0); p(1,3)=Hull(-0.2); p(1,4)=Hull( 2); 
    p(2,1)=Hull(0.00); p(2,2)=Hull(0); p(2,3)=Hull( 0.5); p(2,4)=Hull(-5); 
  }
  }
  if ( System == "sSystem") {
    if ( comp == 1 ) {
      p(1,3)=Hull(-0.8); p(1,5)=Hull(+12);
      p(2,1)=Hull(+0.5); p(2,5)=Hull(-10);
    } 
    if ( comp == 2 ) {
      p(1,1)=Hull(+0.5);  p(1,5)=Hull(+ 8);
      p(2,2)=Hull(+0.75); p(2,5)=Hull(- 3);
    } 
    if ( comp == 3 ) {
      p(1,2)=Hull(+0.75); p(1,5)=Hull(+ 3);
      p(2,3)=Hull(+0.5);  p(2,4)=Hull(+0.2); p(2,5)=Hull(- 5);
    } 
    if ( comp == 4 ) {
      p(1,1)=Hull(+0.5); p(1,5)=Hull(+ 2);
      p(2,4)=Hull(+0.8); p(2,5)=Hull(- 6);
    }    
  }
}  

////////////////////////////////////////////////////////////////////////////
//
void widen(IM &p, int comp, double rel, string System)
{ 
  p *= (1 + rel*Hull(-1,+1));

  // When we solve for the first component parameters of the GMA example, 
  // we use the fact that the other parameters have already been 
  // correctly reconstructed to an accuracy of +- 2.4e-4.
  if ( (System == "GMA") && (comp == 1) ) {
    INTERVAL EPS = 2.4e-4*Hull(-1,+1); 
    p(2,1)=Hull(0.50)+EPS; p(2,2)=Hull(-0.1)+EPS; p(2,4)=Hull(-3)+EPS; 
    p(3,1)=Hull(0.75)+EPS; p(3,3)=Hull(-0.2)+EPS; p(3,4)=Hull(-2)+EPS; 
  }
}

////////////////////////////////////////////////////////////////////////////
//
INTERVAL positive(const INTERVAL &x)
{ return INTERVAL(0.0, Sup(x)); }

INTERVAL negative(const INTERVAL &x)
{ return INTERVAL(Inf(x), 0.0); }

////////////////////////////////////////////////////////////////////////////
//
// When used, sets the topology according to the model.
//
void setTopology(IM &p, string System, int comp)
{
  int cDim = ColDimension(p);      // cDim = d + 1.
  if ( System == "GMA") {
    if ( comp == 1 ) {
      p(1,1) = Hull(0); p(1,cDim) = positive(p(1,cDim));
      p(2,3) = Hull(0); p(2,cDim) = negative(p(2,cDim));
      p(3,2) = Hull(0); p(3,cDim) = negative(p(3,cDim));
    }
    if ( comp == 2 ) {
      IM q(p); Resize(p,2,cDim);
      p(1,1) = q(1,1);  p(1,2) = q(1,2); 
      p(1,3) = Hull(0); p(1,cDim) = q(1,cDim);
      p(2,1) = Hull(0); p(2,2) = q(2,2); 
      p(2,3) = Hull(0); p(2,cDim) = q(2,cDim);
    }  
    if ( comp == 3 ) {
      IM q(p); Resize(p,2,cDim);
      p(1,1) = q(1,1);  p(1,2) = Hull(0); 
      p(1,3) = q(1,3);  p(1,cDim) = q(1,cDim); 
      p(2,1) = Hull(0); p(2,2) = Hull(0); 
      p(2,3) = q(2,3);  p(2,cDim) = q(2,cDim);
    } 
  }
  else if ( System == "sSystem") {
    if ( comp == 1 ) {
      p(1,cDim) = positive(p(1,cDim)); p(2,cDim) = negative(p(2,cDim));
      p(1,1) = Hull(0); p(1,2) = Hull(0); p(1,4) = Hull(0);
      p(2,2) = Hull(0); p(2,3) = Hull(0); p(2,4) = Hull(0);
    }
    if ( comp == 2 ) {
      p(1,cDim) = positive(p(1,cDim)); p(2,cDim) = negative(p(2,cDim));
      p(1,2) = Hull(0); p(1,3) = Hull(0); p(1,4) = Hull(0);
      p(2,1) = Hull(0); p(2,3) = Hull(0); p(2,4) = Hull(0);
    }  
    if ( comp == 3 ) {
      p(1,cDim) = positive(p(1,cDim)); p(2,cDim) = negative(p(2,cDim));
      p(1,1) = Hull(0); p(1,3) = Hull(0); p(1,4) = Hull(0);
      p(2,1) = Hull(0); p(2,2) = Hull(0); 
    } 
    if ( comp == 4 ) {
      p(1,cDim) = positive(p(1,cDim)); p(2,cDim) = negative(p(2,cDim));
      p(1,2) = Hull(0); p(1,3) = Hull(0); p(1,4) = Hull(0);
      p(2,1) = Hull(0); p(2,2) = Hull(0); p(2,3) = Hull(0);
    }
  }
  else {
    cout << "Undefined topology name. Quitting!" << endl;
    exit(1);
  }
}

////////////////////////////////////////////////////////////////////////////
//
bool findSolution(IM const &parameterDomain, const IM &dataSet, 
		  int Comp, double Tol)
{
  List<IM> resultList;
  IM dataSlice(RowDimension(dataSet), ColDimension(dataSet)/2 + 2);
  getDataSlice(dataSlice, dataSet, Comp);
  
  reconstructParameters(resultList, dataSlice, parameterDomain, Tol);
    
  if ( IsEmpty(resultList) ) {
    cout << "No solutions found." << endl;
    return false;
  }
  cout << "|resultList| = " << Length(resultList) << endl;
  cout << "[resultList] = " << endl; 
  prettyPrint(Hull(resultList));
  cout << "<resultList> = " << endl; 
  prettyPrint(averageOfTheList(resultList));
  return true;
}

////////////////////////////////////////////////////////////////////////////
//
int main(int argc, char * argv[])
{
  if ( (argc < 5) || (6 < argc) ) {
    cerr << "Syntax: gma datadir dim tol comp [rel]" << endl;
    exit(0);
  }
  cout << endl
       << "***********************************************************" << endl
       << "***** CONSTRAINT PROPAGATION PARAMETER RECONSTRUCTION *****" << endl
       << endl << "Latest compilation: " << __DATE__ << ", " << __TIME__ 
       << endl << endl;
  
  // initialize random generator
  // srand ( time(NULL) );
  
  cout.precision(3);
  string dataDir(argv[1]);         // Name of the data directory.
  int    Dim (atoi(argv[2]));      // Number of variables.
  double Tol (atof(argv[3]));      // Stopping tolerance.
  int    Comp(atoi(argv[4]));      // The component to reconstruct.
  string System = "GMA";           // The actual system under study.

  int Samp, Runs;
  if ( !getArguments(Samp, Runs, Dim, dataDir) ) {
    cout << "Could not read data from '" << dataDir << ". Quitting!" << endl;
    exit(1);
  }

  int gmaTerms = 3;
  if ( dataDir == "voit") {
    System = "sSystem";  
    gmaTerms = 2;
  }
  cout << "Dimension : " << Dim  << endl
       << "Samples   : " << Samp << endl
       << "Runs      : " << Runs << endl
       << "Tolerance : " << Tol  << endl
       << "Data dir  : " << dataDir << endl
       << "Component : " << Comp << endl
       << "ODE-system: " << System << endl;

  IM globalParameterDomain(gmaTerms, Dim + 1);  
  Initialize(globalParameterDomain, Hull(0));
  if ( argc == 6 ) {
    double Rel(atof(argv[5]));     // The relative widening of the domain.
    cout << "Relative  : " << 100*Rel << "%" << endl;
    setTargetParameters(globalParameterDomain, System, Comp);
    widen(globalParameterDomain, Comp, Rel, System);
  }
  else {
    setParameterDomain(globalParameterDomain, System);
    setTopology(globalParameterDomain, System, Comp);
  }
  cout << "Global parameterDomain:" << endl;
  prettyPrint(globalParameterDomain);
  
  IM dataSet(Samp*Runs, 2*Dim);    // (x_i1,...,x_id, s_i1,...,s_id).
  readData(dataSet, Dim, Samp, Runs, dataDir);
#ifdef USE_SLOPE_CONSTRAINTS
  cout << "Using the slope constraints." << endl;
#endif
#ifdef USE_GENERAL_CONSTRAINTS
  cout << "Using the general constraints." << endl;
#endif
  clock(START_TIMING);  
  findSolution(globalParameterDomain, dataSet, Comp, Tol);
  clock(STOP_TIMING);  

  return 0;
}

////////////////////////////////////////////////////////////////////////////








