/*   File: utilities.cc

     Various utility functions used by the program 'cpParam'. These functions
     are completely s-system independent.

     Author: Warwick Tucker <warwick@math.uu.se>
     Latest edit: Thu Oct 27 14:04:16 CEST 2005
*/

#include "utilities.h"

////////////////////////////////////////////////////////////////////////////
// 'START_TIMING' starts the clock, 'STOP_TIMING' stops the clock,
// and prints the elapsed time in hh:mm:ss. 'SHOW_TIMING' prints the 
// elapsed time.
//
void clock(const command &cmd)
{
  struct tm     *ptr;          // Used for timing the program 
  static time_t  time_start;   // start time 

  if ( cmd == START_TIMING ) {
    time_start = time(NULL);
    ptr = localtime(&time_start); 
    cout << "Started computations: " << asctime(ptr);  
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;  
    
    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
      cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: "; 
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) { seconds -= 60; minutes++; }
    while ( minutes >= 60 ) { minutes -= 60;   hours++; }
    while (   hours >= 24 ) {   hours -= 24;    days++; }
    while (    days >=  7 ) {    days -=  7;   weeks++; }
    if ( weeks > 0)
      cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
      cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
      cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
      cout << minutes << " minute" << (minutes != 1 ? "s, " : ", "); 
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl; 
  }
}

////////////////////////////////////////////////////////////////////////////
//
//  Returns a random number from [min, max].
//
double uniformRnd(double min, double max)
{
  return ((double) rand() / RAND_MAX) * (max - min) + min;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the maximal component diameter of the matrix..
//
double Diameter(const IM &M)
{
  double result = Diam(M(1,1));
  for (int i = 1; i <= RowDimension(M); i++) 
    for (int j = 1; j <= ColDimension(M); j++) {
      double dM = Diam(M(i,j));
      if (result < dM )
	result = dM;
    }
  return result;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the hull of all list members.
//
IM Hull(List<IM> &boxList)
{
  List<IM> dummyList;
  IM result = First(boxList);
  Next(boxList);
  while ( !Finished(boxList) ) {
    result = Hull(result, Current(boxList));
    Next(boxList);
  }
  return result;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the hull of all list members.
//
BOX Hull(List<BOX> &boxList)
{
  List<BOX> dummyList;
  BOX result = First(boxList);
  Next(boxList);
  while ( !Finished(boxList) ) {
    result = Hull(result, Current(boxList));
    Next(boxList);
  }
  return result;
}

/////////////////////////////////////////////////////////////////////
//
// Returns the average Mid(b_1 + ... + b_N)/N.
//
MATRIX averageOfTheList(List<IM> &matList)
{
  int listN     = Length(matList);
  MATRIX outMat = Mid(First(matList)); Next(matList);
  while (!Finished(matList) ) {
    outMat += Mid(Current(matList));
    Next(matList);
  }
  return outMat / listN;
}

/////////////////////////////////////////////////////////////////////
//
// Returns the average Mid(b_1 + ... + b_N)/N.
//
VECTOR averageOfTheList(List<BOX> &boxList)
{
  int listN     = Length(boxList);
  VECTOR outBox = Mid(First(boxList)); Next(boxList);
  while (!Finished(boxList) ) {
    outBox += Mid(Current(boxList));
    Next(boxList);
  }
  return outBox / listN;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the index of the widest component of p.
//
void widestElement(int &iDom, int &jDom, const IM &p)
{
  double diam = Diam(p(1,1));
  iDom = 1; 
  jDom = 1;
  for (int i = 1; i <= RowDimension(p); i++) 
    for (int j = 1; j <= ColDimension(p); j++)  {
      double ijDiam = Diam(p(i,j));
      if ( diam < ijDiam ) {
	diam = ijDiam;
	iDom = i; jDom = j;
      }
    }
}

////////////////////////////////////////////////////////////////////////////
//
// Splits (and stores) the box p across its widest component.
//
void splitAndStore(const IM &p, List<IM> &ioList)
{
  IM  loP = p;
  IM  hiP = p;
  int i,j;
  widestElement(i, j, p);

  loP(i,j) = INTERVAL(Inf(p(i,j)), Mid(p(i,j)));
  hiP(i,j) = INTERVAL(Mid(p(i,j)), Sup(p(i,j)));

  ioList += loP;
  ioList += hiP;
}

////////////////////////////////////////////////////////////////////////////
//
// Returns the index of the widest component of p.
//
int widestElement(const BOX &p)
{
  int    pDim = Dimension(p);
  double diam = Diam(p(1));
  int    iDom = 1; 
  for (int i = 2; i <= pDim; i++) {
    double iDiam = Diam(p(i));
    if ( diam < iDiam ) {
      diam = iDiam;
      iDom = i;
    }
  }
  return iDom;
}

////////////////////////////////////////////////////////////////////////////
//
// Splits (and stores) the box p across its widest component.
//
void splitAndStore(const BOX &p, List<BOX> &ioList)
{
  BOX  loP = p;
  BOX  hiP = p;
  int    i = widestElement(p);

  loP(i) = INTERVAL(Inf(p(i)), Mid(p(i)));
  hiP(i) = INTERVAL(Mid(p(i)), Sup(p(i)));

  ioList += loP;
  ioList += hiP;
}

////////////////////////////////////////////////////////////////////////////
//
// Interlaces all data to get early large variation in phase space.
//
void insertData(IM &dS, const IM &xVal, const IM &sVal, int run, int runs)
{
  int samp = RowDimension(xVal);
  int dim  = ColDimension(xVal);
  for (int i = 1; i <= samp; i++)
    for (int j = 1; j <= dim; j++) {
      dS(run + (i-1)*runs, j)     = xVal(i,j);
      dS(run + (i-1)*runs, j+dim) = sVal(i,j);
    }
}

////////////////////////////////////////////////////////////////////////////
//
// Reads thin data and converts it to intervals stored in a matrix.
//
IM getData(istream& inFile, int dim, int samp)
{
  IM data(samp, dim);
  for (int i = 1; i <= samp; i++)
    for (int j = 1; j <= dim; j++) {
      double inputDbl;
      inFile >> inputDbl;
      data(i,j) = Hull(inputDbl);
    }
#ifdef W_DEBUG
  cout << "Data..." << endl << data << endl;    
#endif
  return data;
}

////////////////////////////////////////////////////////////////////////////
//
string int2string(int number)
{
   ostringstream oss;
   oss << number;
   return oss.str();
}
 
////////////////////////////////////////////////////////////////////////////
//
// Opens and extracts data from the files 'xValuei.txt' and 'sValuei.txt'
// where i = 1,...,runs.
//
void readData(IM &dS, int dim, int samples, int runs, string dataDir)
{
  string beginning = "data/";
  string ending    = ".txt";
  for (int i = 1; i <= runs; i++) {
    IM xVal, sVal;
    
    // Construct the various file names.
    string xValFileName = beginning + dataDir + "/xValue";
    string sValFileName = beginning + dataDir + "/sValue";
    xValFileName += int2string(i) + ending;
    sValFileName += int2string(i) + ending;

    // Try to open the files.
    ifstream xValFile(xValFileName.c_str(), ios_base::in);
    if ( !xValFile ) {
      cout << "Cannot open file: " << xValFileName << " for reading" << endl;
      exit(1);
    }
    else {
      //cout << "Opened " << xValFileName << " for reading." << endl;
      xVal = getData(xValFile, dim, samples);
    }
    ifstream sValFile(sValFileName.c_str(), ios_base::in);
    if ( !sValFile ) {
      cout << "Cannot open file: " << sValFileName << " for reading" << endl;
      exit(1);
    }
    else {
      //cout << "Opened " << sValFileName << " for reading." << endl;
      sVal = getData(sValFile, dim, samples);
    }
    insertData(dS, xVal, sVal, i, runs);
  }
  cout << endl;
}

////////////////////////////////////////////////////////////////////////////
//
// [x_1,...,x_d, s_1,...,s_d] --> [x_1,...,x_d, x_i, s_i].
//          dataSet           -->         dataSlice
//
void getDataSlice(IM &dataSlice, const IM &dataSet, int comp)
{
  int Dim = ColDimension(dataSlice) - 2;
  for (int i = 1; i <= Dim; i++) 
    SetCol(dataSlice, i, Col(dataSet,i));            // Copy all phase space.
  SetCol(dataSlice, Dim + 1, Col(dataSet,comp));     // Copy component 'comp'
  SetCol(dataSlice, Dim + 2, Col(dataSet,comp+Dim)); // and its slope data.
}

////////////////////////////////////////////////////////////////////////////
//
// Uses program flags to compute 'Samp' (number of sample points per run)
// and 'Runs' (number of performed experiments).
//
bool getArguments(int &Samp, int &Runs, int Dim, string dataDir) 
{
  string beginning = "data/" + dataDir + "/xValue";
  Runs = 0;
  bool moreRuns = true;
  while ( moreRuns ) {
    string xValFileName = beginning + int2string(Runs+1) + ".txt";
    ifstream xValFile(xValFileName.c_str(), ios_base::in);
    if ( xValFile )
      Runs++;
    else
      moreRuns = false;
    xValFile.close();
  } 
  if ( Runs == 0 )
    return false;
  
  Samp = 0;
  string xValFileName = beginning + int2string(Runs) + ".txt";
  ifstream xValFile(xValFileName.c_str(), ios_base::in);
  VECTOR dataLine(Dim);
  while ( xValFile >> dataLine )
    Samp++;
  if ( Samp == 0 )
    return false;

  return true;
}

////////////////////////////////////////////////////////////////////////////

void prettyPrint(const MATRIX &p)
{
  int Dim = ColDimension(p) - 1;
  int gmaTerms = RowDimension(p);

  for (int i = 1; i <= gmaTerms; i++) {
    cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
    cout << "a(" << i << ")   = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i, Dim + 1) << endl;
    for (int j = 1; j <= Dim; j++) {
      if ( p(i,j) != 0.0 ) {
	cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
	cout << "g(" << i << "," << j << ") = ";
	cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
	cout << p(i,j) << endl;
      }
    }
  }
  cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
}

////////////////////////////////////////////////////////////////////////////

void prettyPrint(const IM &p)
{
  int Dim = ColDimension(p) - 1;
  int gmaTerms = RowDimension(p);

  for (int i = 1; i <= gmaTerms; i++) {
    cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
    cout << "a(" << i << ")   = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i, Dim + 1)  << "   Diam = " << Diam(p(i, Dim + 1)) 
	 << "   Mid = " << Mid(p(i, Dim + 1)) << endl;
    for (int j = 1; j <= Dim; j++) {
      if ( Diam(p(i,j)) > 0.0 ) {
	cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
	cout << "g(" << i << "," << j << ") = ";
	cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
	cout << p(i,j) << "   Diam = " << Diam(p(i,j))  
	     << "   Mid = " << Mid(p(i,j)) << endl;
      }
    }
  }
  cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
}

////////////////////////////////////////////////////////////////////////////

void prettyPrint(const BOX &p)
{
  int Dim = Dimension(p)/2 - 1;
  cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
  cout << "a    = " << p(1) << "\t"
       << "b    = " << p(Dim+2) << endl;
  for (int i = 1; i <= Dim; i++) {
    cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
    cout << "g(" << i << ") = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i+1) << "\t";
    cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
    cout << "h(" << i << ") = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i+2+Dim) << endl;
  }
  cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
}

////////////////////////////////////////////////////////////////////////////

void prettyPrint(const VECTOR &p)
{
  int Dim = Dimension(p)/2 - 1;
  cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
  cout << "a    = " << p(1) << "\t"
       << "b    = " << p(Dim+2) << endl;
  for (int i = 1; i <= Dim; i++) {
    cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
    cout << "g(" << i << ") = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i+1) << "\t";
    cout.unsetf(ios_base::showpos); cout.unsetf(ios_base::scientific);
    cout << "h(" << i << ") = ";
    cout.setf(ios_base::showpos);   cout.setf(ios_base::scientific);
    cout << p(i+2+Dim) << endl;
  }
  cout.unsetf(ios_base::showpos);   cout.unsetf(ios_base::scientific);
}

////////////////////////////////////////////////////////////////////////////
